import pytest

from datetime import datetime, timedelta
from io import BytesIO
from onegov.file import File
from onegov.form import CompleteFormSubmission
from onegov.form import FormCollection
from onegov.form import parse_form
from onegov.form import PendingFormSubmission
from onegov.form.errors import UnableToComplete
from onegov.form.models import hash_definition
from sedate import utcnow
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import FlushError
from textwrap import dedent
from webob.multidict import MultiDict
from werkzeug.datastructures import FileMultiDict
from wtforms.csrf.core import CSRF


def test_form_checksum():
    assert hash_definition('abc') == '900150983cd24fb0d6963f7d28e17f72'


def test_add_form(session):
    collection = FormCollection(session)

    form = collection.definitions.add('Tax Form', definition=dedent("""
        First Name * = ___
        Last Name * = ___
    """))

    assert form.name == 'tax-form'
    assert form.form_class._source == form.definition

    form = collection.definitions.by_name('tax-form')

    assert form.name == 'tax-form'
    assert form.form_class._source == form.definition

    with pytest.raises(FlushError):
        form = collection.definitions.add('Tax Form', definition=dedent("""
            First Name * = ___
            Last Name * = ___
        """))

    assert form.checksum and form.checksum == hash_definition(form.definition)


def test_submit_form(session):
    collection = FormCollection(session)

    form = collection.definitions.add('TPS Report', definition=dedent("""
        First Name * = ___
        Last Name * = ___
        Date = YYYY.MM.DD
    """))

    data = MultiDict([
        ('first_name', 'Bill'),
        ('last_name', 'Lumbergh'),
        ('date', '2011-01-01')
    ])

    submitted_form = form.form_class(data)
    collection.submissions.add('tps-report', submitted_form, state='complete')

    form = collection.definitions.by_name('tps-report')
    submission = form.submissions[0]

    assert submission.checksum and submission.checksum == hash_definition(
        form.definition) == hash_definition(submission.definition)

    stored_form = submission.form_class(data=submission.data)

    assert submitted_form.data == stored_form.data


def test_submission_extra_data(session):
    collection = FormCollection(session)

    form = collection.definitions.add('TPS Report', definition=dedent("""
        First Name * = ___
        Last Name * = ___
        E-Mail = @@@
        Date = YYYY.MM.DD
    """))

    data = MultiDict([
        ('first_name', 'Bill'),
        ('last_name', 'Lumbergh'),
        ('e_mail', 'bill.lumbergh@initech.com'),
        ('date', '2011-01-01')
    ])

    submitted_form = form.form_class(data)
    submission = collection.submissions.add(
        'tps-report', submitted_form, state='complete')

    assert submission.title == 'Bill, Lumbergh'
    assert submission.email == 'bill.lumbergh@initech.com'


def test_definitions_with_submissions_count(session):
    collection = FormCollection(session)

    form = collection.definitions.add('Newsletter', definition="E-Mail = @@@")

    data = MultiDict([
        ('e_mail', 'test@example.org'),
    ])

    s1 = collection.submissions.add(
        'newsletter', form.form_class(data), state='complete')

    form = next(collection.get_definitions_with_submission_count())
    assert form.submissions_count == 1

    s2 = collection.submissions.add(
        'newsletter', form.form_class(data), state='complete')

    form = next(collection.get_definitions_with_submission_count())
    assert form.submissions_count == 2

    collection.submissions.delete(s1)

    form = next(collection.get_definitions_with_submission_count())
    assert form.submissions_count == 1

    collection.submissions.delete(s2)

    form = next(collection.get_definitions_with_submission_count())
    assert form.submissions_count == 0


def test_submit_pending(session):
    collection = FormCollection(session)

    form = collection.definitions.add('tweet', definition=dedent("""
        handle * = ___
        tweet * = ___[140]
    """))

    data = MultiDict([
        ('handle', '@href'),
        ('tweet', (
            "I think I found a way to tweet more than 140 characters! "
            "Finally I can tell my life's story in a single tweet. "
            "#hackersgonnahack #infosec #yolo"
        ))
    ])

    # pending forms may be submitted even if the are not valid
    submitted_form = form.form_class(data)
    assert not submitted_form.validate()

    submission = collection.submissions.add(
        'tweet', submitted_form, state='pending')
    assert isinstance(submission, PendingFormSubmission)

    assert 'handle' in submission.data
    assert 'tweet' in submission.data

    with pytest.raises(UnableToComplete):
        collection.submissions.complete_submission(submission)

    submission.data['tweet'] = "Nevermind, it didn't work #mybad"
    collection.submissions.complete_submission(submission)

    submission = collection.submissions.by_state('complete').first()
    submission.state == 'complete'
    submission.__class__ == CompleteFormSubmission


def test_remove_old_pending_submissions(session):
    collection = FormCollection(session)

    signup = collection.definitions.add('Signup', definition="E-Mail = @@@")

    data = MultiDict([('e_mail', 'info@example.org')])
    form = signup.form_class(data)

    collection.submissions.add('signup', form, state='complete')
    collection.submissions.add('signup', form, state='pending')

    assert collection.submissions.query().count() == 2

    collection.submissions.remove_old_pending_submissions(
        datetime.utcnow() - timedelta(hours=1))

    assert collection.submissions.query().count() == 2

    collection.submissions.remove_old_pending_submissions(
        datetime.utcnow() + timedelta(hours=1))

    assert collection.submissions.query().count() == 1


def test_no_store_csrf_token(session):
    collection = FormCollection(session)

    class MockCSRF(CSRF):

        def generate_csrf_token(self, csrf_token_field):
            return '0xdeadbeef'

    signup = collection.definitions.add('Signup', definition="E-Mail = @@@")

    data = MultiDict([
        ('e_mail', 'info@example.org'),
        ('csrf_token', '0xdeadbeef')
    ])

    form = signup.form_class(data, meta=dict(csrf=True, csrf_class=MockCSRF))
    submission = collection.submissions.add('signup', form, state='complete')

    assert 'e_mail' in submission.data
    assert 'csrf_token' not in submission.data


def test_delete_without_submissions(session):
    collection = FormCollection(session)

    form = collection.definitions.add('Newsletter', definition="E-Mail *= @@@")
    data = MultiDict([('e_mail', 'billg@microsoft.com')])

    collection.submissions.add(
        'newsletter', form.form_class(data), state='complete')

    with pytest.raises(IntegrityError):
        collection.definitions.delete('newsletter')


def test_delete_with_submissions(session):
    collection = FormCollection(session)

    form = collection.definitions.add('Newsletter', definition="E-Mail *= @@@")
    data = MultiDict([('e_mail', 'billg@microsoft.com')])

    collection.submissions.add(
        'newsletter', form.form_class(data), state='complete')
    collection.definitions.delete('newsletter', with_submissions=True)

    assert collection.submissions.query().count() == 0
    assert collection.definitions.query().count() == 0


def test_delete_with_pending_submissions(session):
    collection = FormCollection(session)

    form = collection.definitions.add('Newsletter', definition="E-Mail *= @@@")
    data = MultiDict([('e_mail', 'billg@microsoft.com')])

    collection.submissions.add(
        'newsletter', form.form_class(data), state='pending')
    collection.definitions.delete('newsletter', with_submissions=False)

    assert collection.submissions.query().count() == 0
    assert collection.definitions.query().count() == 0


def test_delete_fail_with_submissions(session):
    collection = FormCollection(session)

    form = collection.definitions.add('Newsletter', definition="E-Mail *= @@@")
    data = MultiDict([('e_mail', 'billg@microsoft.com')])

    collection.submissions.add(
        'newsletter', form.form_class(data), state='complete')

    with pytest.raises(IntegrityError):
        collection.definitions.delete('newsletter', with_submissions=False)


def test_file_submissions_update(session):
    collection = FormCollection(session)

    # upload a new file
    definition = collection.definitions.add('File', definition="File = *.txt")

    data = FileMultiDict()
    data.add_file('file', BytesIO(b'foobar'), filename='foobar.txt')

    submission = collection.submissions.add(
        'file', definition.form_class(data), state='pending')

    assert len(submission.files) == 1
    assert submission.files[0].checksum == '3858f62230ac3c915f300c664312c63f'

    # replace the existing file
    previous_file = submission.files[0]
    session.refresh(submission)

    data = FileMultiDict()
    data.add('file', 'replace')
    data.add_file('file', BytesIO(b'barfoo'), filename='foobar.txt')

    collection.submissions.update(submission, definition.form_class(data))
    session.flush()
    session.refresh(submission)

    assert len(submission.files) == 1
    assert previous_file.id != submission.files[0].id
    assert previous_file.checksum != submission.files[0].checksum

    # keep the file
    previous_file = submission.files[0]
    session.refresh(submission)

    data = FileMultiDict()
    data.add('file', 'keep')
    data.add_file('file', BytesIO(b''), filename='foobar.txt')

    collection.submissions.update(submission, definition.form_class(data))
    session.flush()
    session.refresh(submission)

    assert len(submission.files) == 1
    assert previous_file.id == submission.files[0].id
    assert previous_file.checksum == submission.files[0].checksum

    # delete the file
    session.refresh(submission)

    data = FileMultiDict()
    data.add('file', 'delete')
    data.add_file('file', BytesIO(b'asdf'), filename='foobar.txt')

    collection.submissions.update(submission, definition.form_class(data))
    assert len(submission.files) == 0


def test_file_submissions_cascade(session):

    collection = FormCollection(session)

    # upload a new file
    definition = collection.definitions.add('File', definition="File = *.txt")

    data = FileMultiDict()
    data.add_file('file', BytesIO(b'foobar'), filename='foobar.txt')

    collection.submissions.add(
        'file', definition.form_class(data), state='pending')

    assert session.query(File).count() == 1
    session.flush()

    collection.submissions.remove_old_pending_submissions(older_than=(
        datetime.utcnow() + timedelta(seconds=60)))

    session.flush()

    assert session.query(File).count() == 0


def test_get_current(session):
    collection = FormCollection(session)

    form = collection.definitions.add('Newsletter', definition="E-Mail *= @@@")
    data = MultiDict([('e_mail', 'billg@microsoft.com')])

    submission = collection.submissions.add(
        'newsletter', form.form_class(data), state='complete')

    submission.created = submission.modified = utcnow() - timedelta(days=1)

    assert not collection.submissions.by_id(submission.id, current_only=True)
    assert collection.submissions.by_id(submission.id, current_only=False)


def test_add_externally_defined_submission(session):
    collection = FormCollection(session)

    form_class = parse_form("E-Mail *= @@@")
    form = form_class(data={'e_mail': 'info@example.org'})

    submission = collection.submissions.add_external(form, state='pending')

    assert collection.definitions.query().count() == 0
    assert collection.submissions.query().count() == 1

    stored_form = submission.form_class(data=submission.data)

    assert stored_form.e_mail.data == 'info@example.org'
    assert stored_form.e_mail.data == form.e_mail.data

    # externally defined submission are not automatically removed
    date = datetime.utcnow() + timedelta(seconds=60)

    collection.submissions.remove_old_pending_submissions(older_than=date)
    assert collection.submissions.query().count() == 1

    collection.submissions.remove_old_pending_submissions(
        older_than=date, include_external=True)
    assert collection.submissions.query().count() == 0



from more.webassets import WebassetsApp
from onegov.core.security import Public
from onegov.form import _
from onegov.form.parser.core import flatten_fieldsets, parse_formcode
from onegov.form.parser.snippets import Snippets
from onegov.form.errors import FormError


class FormApp(WebassetsApp):
    pass


class FormcodeParseFields(object):
    pass


@FormApp.path(path='formcode-snippets', model=Snippets)
def get_formcode_snippets():
    return Snippets()


@FormApp.path(path='formcode-fields', model=FormcodeParseFields)
def get_formcode_parse_fields():
    return FormcodeParseFields()


@FormApp.json(model=Snippets, permission=Public)
def view_formcode_snippets(self, request):
    return {
        'labels': {
            'required': request.translate(_("Required")),
            'optional': request.translate(_("Optional")),
        },
        'snippets': tuple(self.translated(request))
    }


@FormApp.json(model=FormcodeParseFields, permission=Public,
              request_method='GET')
@FormApp.json(model=FormcodeParseFields, permission=Public,
              request_method='POST')
def view_parse_formcode(self, request):
    formcode = request.params.get('formcode') or request.text

    if not formcode:
        return {}

    try:
        return [
            {
                'id': field.id,
                'human_id': field.human_id,
                'type': field.type,
            }
            for field in flatten_fieldsets(parse_formcode(formcode))
        ]
    except (FormError, AttributeError, TypeError):
        return {'error': True}

    return []


@FormApp.webasset_path()
def get_js_path():
    return 'assets/js'


@FormApp.webasset('formcode')
def get_formcode_asset():
    yield 'utils.js'
    yield 'watcher.jsx'
    yield 'togglebutton.jsx'
    yield 'snippets.jsx'
    yield 'format.jsx'
    yield 'select.jsx'



import logging
log = logging.getLogger('onegov.form')  # noqa
log.addHandler(logging.NullHandler())  # noqa

from translationstring import TranslationStringFactory
_ = TranslationStringFactory('onegov.form')  # noqa

from onegov.form.collection import (
    FormCollection,
    FormSubmissionCollection,
    FormDefinitionCollection
)
from onegov.form.core import (
    FieldDependency,
    Form,
    merge_forms,
    move_fields,
    with_options,
)
from onegov.form.display import render_field
from onegov.form.integration import FormApp
from onegov.form.models import (
    FormDefinition,
    FormFile,
    FormSubmission,
    PendingFormSubmission,
    CompleteFormSubmission
)
from onegov.form.parser import find_field
from onegov.form.parser import flatten_fieldsets
from onegov.form.parser import parse_form
from onegov.form.parser import parse_formcode
from onegov.form.parser import WTFormsClassBuilder
from onegov.form.utils import decimal_range, as_internal_id

__all__ = [
    'as_internal_id',
    'CompleteFormSubmission',
    'decimal_range',
    'find_field',
    'flatten_fieldsets',
    'FieldDependency',
    'Form',
    'FormApp',
    'FormCollection',
    'FormDefinition',
    'FormDefinitionCollection',
    'FormFile',
    'FormSubmission',
    'FormSubmissionCollection',
    'merge_forms',
    'move_fields',
    'parse_form',
    'parse_formcode',
    'PendingFormSubmission',
    'render_field',
    'with_options',
    'WTFormsClassBuilder',
]



import re
import tempfile

from cgi import FieldStorage
from gzip import GzipFile
from io import BytesIO
from onegov.core.utils import dictionary_to_binary
from onegov.form import Form
from onegov.form.fields import HtmlField
from onegov.form.fields import MultiCheckboxField
from onegov.form.fields import OrderedMultiCheckboxField
from onegov.form.fields import UploadField


def create_file(mimetype, filename, content):
    fs = FieldStorage()
    fs.file = tempfile.TemporaryFile("wb+")
    fs.type = mimetype
    fs.filename = filename
    fs.file.write(content)
    fs.file.seek(0)
    return fs


def test_upload_file():
    field = UploadField()
    field = field.bind(Form(), 'upload')

    textfile = create_file('text/plain', 'test.txt', b'foobar')
    data = field.process_fieldstorage(textfile)

    assert data['filename'] == 'test.txt'
    assert data['mimetype'] == 'text/plain'
    assert data['size']
    assert data['data']

    def decompress(data):
        with GzipFile(filename='', mode='r', fileobj=BytesIO(data)) as f:
            return f.read()

    assert dictionary_to_binary(data) == b'foobar'


def test_ordered_multi_checkbox_field():
    ordinary = MultiCheckboxField(choices=[
        ('c', 'C'),
        ('b', 'B'),
        ('a', 'A')
    ])
    ordered = OrderedMultiCheckboxField(choices=[
        ('c', 'C'),
        ('b', 'B'),
        ('a', 'A')
    ])
    ordinary = ordinary.bind(Form(), 'choices')
    ordered = ordered.bind(Form(), 'choices')

    ordinary.data = ordered.data = []

    assert re.findall(r'value="((a|b|c){1})"', ordinary()) == [
        ('c', 'c'),
        ('b', 'b'),
        ('a', 'a'),
    ]

    assert re.findall(r'value="((a|b|c){1})"', ordered()) == [
        ('a', 'a'),
        ('b', 'b'),
        ('c', 'c'),
    ]


def test_html_field():
    form = Form()
    field = HtmlField()
    field = field.bind(form, 'html')

    field.data = ''
    assert 'class="editor"' in field()
    assert field.validate(form)

    field.data = '<script>alert(0)</script>'
    field.validate(form)
    assert '<script>' not in field.data



""" onegov.form includes it's own markdownish form syntax, inspired by
https://github.com/maleldil/wmd

The goal of this syntax is to enable the creation of forms through the web,
without having to use javascript, html or python code.

Also, just like Markdown, we want this syntax to be readable by humans.

How it works
============

Internally, the form syntax is converted into a YAML file, which is in turn
parsed and turned into a WTForms class. We decided to go for the intermediate
YAML file because it's easy to define a Syntax which correctly supports
indentation. Our pyparsing approach was flimsy at best.

Parser Errors
=============

There's currently no sophisticated error check. It's possible that the parser
misunderstand something you defined without warning. So be careful to check
that what you wanted was actually what you got.

Syntax
======

Fields
------

Every field is identified by a label, an optional 'required' indicator and a
field definition. The Label can be any kind of text, not including ``*`` and
``=``.
The ``*`` indicates that a field is required. The ``=`` separates the
identifier from the definition.

A required field starts like this::

    My required field * =

An optional field starts like this::

    My optional field =

Following the identifier is the field definition. For example, this defines
a textfield::

    My textfield = ___

All possible fields are documented further below.

Fieldsets
---------

Fields are grouped into fieldsets. The fieldset of a field is the fieldset
that was last defined::

    # Fieldset 1
    I belong to Fieldset 1 = ___

    # Fieldset 2
    I belong to Fieldset 2 = ___

If no fieldset is defined, the fields don't belong to a fieldset. To stop
putting fields in a fieldset, define an empty fieldeset::

    # Fieldset 1
    I belong to Fieldset 1 = ___

    # ...
    I don't belong to a Fieldset = ___

Available Fields
----------------

Textfield
~~~~~~~~~

A textfield consists of exactly three underscores::

    I'm a textfield = ___

If the textfield is limited in length, the length can be given::

    I'm a limited textfield = ___[50]

The length of such textfields is validated.

Additionally, textfields may use regexes to validate their contents::

    I'm a numbers-only textfield = ___/^[0-9]+$

You *can* combine the length with a regex, though you probably don't want to::

    I'm a length-limited numbers-only textfield = ___[4]/^[0-9]+$

This could be simplified as follows:

    I'm a length-limited numbers-only textfield = ___/^[0-9]{0,4}$

Note that you don't *need* to specify the beginning (^) and the end ($) of the
string, but not doing so might result in unexpected results. For example,
while '123abc' is invalid for ``___/^[0-9]+$``, it is perfectly valid for
``___/[0-9]+``. The latter only demands that the text starts with a number,
not that it only consists of numbers!

Textarea
~~~~~~~~

A textarea has no limit and consists of exactly three dots::

    I'm a textarea = ...

Optionally, the number of rows can be passed to the field. This changes the
way the textarea looks, not the way it acts::

    I'm a textarea with 10 rows = ...[10]

Password
~~~~~~~~

A password field consists of exactly three stars::

    I'm a password = ***

E-Mail
~~~~~~

An e-mail field consists of exactly three ``@``::

    I'm an e-mail field = @@@


URL
~~~

An url field consists of the http/https prefix::

    I'm an url field = http://
    I'm the exact same = https://

Whether or not you enter http or https has no bearing on the validation.

Date
~~~~

A date (without time) is defined by this exact string: ``YYYY.MM.DD``::

    I'm a date field = YYYY.MM.DD

Note that this doesn't mean that the date format can be influenced.

Datetime
~~~~~~~~

A date (with time) is defined by this exact string: ``YYYY.MM.DD HH:MM``::

    I'm a datetime field = YYYY.MM.DD HH:MM

Again, this doesn't mean that the datetime format can be influenced.

Time
~~~~

A Time is defined by this exact string: ``HH:MM``::

    I'm a time field = HH:MM

One more time, this doesn't mean that the datetime format can be influenced.

Numbers
~~~~~~~

There are two types of number fields. An integer and a float field::

    I'm an integer field = 0..99
    I'm an integer field of a different range = -100..100

    I'm a float field = 0.00..99.00
    I'm an float field of a different range = -100.00..100.00

Files
~~~~~

A file upload is defined like this::

    I'm a file upload field = *.*

This particular example would allow any file. To allow only certain files
do something like this::

    I'm a image filed = *.png|*.jpg|*.gif
    I'm a document = *.doc
    I'm any document = *.doc|*.pdf

The files are checked against their file extension. Onegov.form also checks
that uploaded files have the mimetype they claim to have and it won't accept
obviously dangerous uploads like binaries (unless you really want to).

Standard Numbers
~~~~~~~~~~~~~~~~

onegov.form uses `python-stdnum \
<https://github.com/arthurdejong/python-stdnum>`_ to offer a wide range of
standard formats that are guaranteed to be validated.

To use, simply use a `#`, followed by the stdnum format to use::

    I'm a valid IBAN (or empty) = # iban
    I'm a valid IBAN (required) * = # iban

The format string after the `#` must be importable from stdnum. In other words,
this must work, if you are using ``ch.ssn`` (to use an example)::

    $ python
    >>> from stdnum.ch import ssn

This is a bit of an advanced feature and since it delegates most work to an
external library there's no guarantee that a format once used may be reused
in the future.

Still, the library should be somewhat stable and the benefit is huge.

To see the available format, have a look at the docs:
`<http://arthurdejong.org/python-stdnum/doc/1.1/index.html#available-formats>`_

Radio Buttons
~~~~~~~~~~~~~

Radio button fields consist of x radio buttons, out of which one may be
preselected. Radio buttons need to be indented on the lines after the
definition::

    Gender =
        ( ) Female
        ( ) Male
        (x) I don't want to say

Radio buttons also have the ability to define optional form parts. Those
parts are only shown if a question was answered a certain way.

Form parts are properly nested if they lign up with the label above them.

For example::

    Delivery Method =
        ( ) Pickup
            Pickup Time * = ___
        (x) Address
            Street * = ___
            Town * = ___

Here, the street and the town only need to be provided, if the delivery method
is 'Address'. If the user selects a different option, the fields are not
shown and they will not be required.

On the other hand, if 'Pickup' is selected, the 'Pickup Time' needs to be
filled out and the address options are hidden.

This kind of nesting may continue ad infinitum. Meaning you can nest radio
buttons as deeply as you like. Note however, that this is discouraged and that
your users will not be too happy if you present them with a deeply nested
form.

More than one level of nesting is a clear indicator that your form is too
complex.

Checkboxes
~~~~~~~~~~

Checkboxes work exactly like radio buttons, just that you can select
multiple fields::

    Extras =
        [x] Phone insurance
        [ ] Phone case
        [x] Extra battery

Just like radiobuttons, checkboxes may be nested to created dependencies::

    Additional toppings =
        [ ] Salami
        [ ] Olives
        [ ] Other
            Description = ___

Pricing Information
~~~~~~~~~~~~~~~~~~~

Radio buttons and checkboxes may be priced. For example, the following order
form can be modeled::

    Node Size =
        ( ) Small (20 USD)
        (x) Medium (30 USD)
        ( ) Large (40 USD)

    Extras =
        [x] Second IP Address (20 CHF)
        [x] Backup (20 CHF)

The additional pricing metadata can be used to provide simple order forms.
As in any other form, dependencies are taken into account.

"""

import pyparsing as pp
import re
import yaml

from cached_property import cached_property
from dateutil import parser as dateutil_parser
from decimal import Decimal
from onegov.core.utils import Bunch
from onegov.form import errors
from onegov.form.parser.grammar import checkbox
from onegov.form.parser.grammar import date
from onegov.form.parser.grammar import datetime
from onegov.form.parser.grammar import decimal_range_field
from onegov.form.parser.grammar import email
from onegov.form.parser.grammar import field_identifier
from onegov.form.parser.grammar import fieldset_title
from onegov.form.parser.grammar import fileinput
from onegov.form.parser.grammar import integer_range_field
from onegov.form.parser.grammar import password
from onegov.form.parser.grammar import radio
from onegov.form.parser.grammar import stdnum
from onegov.form.parser.grammar import textarea
from onegov.form.parser.grammar import textfield
from onegov.form.parser.grammar import time
from onegov.form.parser.grammar import url
from onegov.form.utils import as_internal_id


# cache the parser elements
def create_parser_elements():
    elements = Bunch()
    elements.identifier = field_identifier()
    elements.fieldset_title = fieldset_title()
    elements.textfield = textfield()
    elements.textarea = textarea()
    elements.password = password()
    elements.email = email()
    elements.url = url()
    elements.stdnum = stdnum()
    elements.datetime = datetime()
    elements.date = date()
    elements.time = time()
    elements.fileinput = fileinput()
    elements.radio = radio()
    elements.checkbox = checkbox()
    elements.integer_range = integer_range_field()
    elements.decimal_range = decimal_range_field()
    elements.boxes = elements.checkbox | elements.radio
    elements.single_line_fields = elements.identifier + pp.MatchFirst([
        elements.textfield,
        elements.textarea,
        elements.password,
        elements.email,
        elements.url,
        elements.stdnum,
        elements.datetime,
        elements.date,
        elements.time,
        elements.fileinput,
        elements.integer_range,
        elements.decimal_range,
    ])

    return elements


ELEMENTS = create_parser_elements()


class CustomLoader(yaml.SafeLoader):
    """ Extends the default yaml loader with customized constructors. """


class constructor(object):
    """ Adds decorated functions to as constructors to the CustomLoader. """

    def __init__(self, tag):
        self.tag = tag

    def __call__(self, fn):
        CustomLoader.add_constructor(self.tag, fn)
        return fn


@constructor('!text')
def construct_textfield(loader, node):
    return ELEMENTS.textfield.parseString(node.value)


@constructor('!textarea')
def construct_textarea(loader, node):
    return ELEMENTS.textarea.parseString(node.value)


@constructor('!email')
def construct_email(loader, node):
    return ELEMENTS.email.parseString(node.value)


@constructor('!url')
def construct_url(loader, node):
    return ELEMENTS.url.parseString(node.value)


@constructor('!stdnum')
def construct_stdnum(loader, node):
    return ELEMENTS.stdnum.parseString(node.value)


@constructor('!date')
def construct_date(loader, node):
    return ELEMENTS.date.parseString(node.value)


@constructor('!datetime')
def construct_datetime(loader, node):
    return ELEMENTS.datetime.parseString(node.value)


@constructor('!time')
def construct_time(loader, node):
    return ELEMENTS.time.parseString(node.value)


@constructor('!radio')
def construct_radio(loader, node):
    return ELEMENTS.radio.parseString(node.value)


@constructor('!checkbox')
def construct_checkbox(loader, node):
    return ELEMENTS.checkbox.parseString(node.value)


@constructor('!fileinput')
def construct_fileinput(loader, node):
    return ELEMENTS.fileinput.parseString(node.value)


@constructor('!password')
def construct_password(loader, node):
    return ELEMENTS.password.parseString(node.value)


@constructor('!decimal_range')
def construct_decimal_range(loader, node):
    return ELEMENTS.decimal_range.parseString(node.value)


@constructor('!integer_range')
def construct_integer_range(loader, node):
    return ELEMENTS.integer_range.parseString(node.value)


def flatten_fieldsets(fieldsets):
    for fieldset in fieldsets:
        yield from flatten_fields(fieldset.fields)


def flatten_fields(fields):
    fields = fields or []

    for field in fields:
        yield field

        if hasattr(field, 'fields'):
            yield from flatten_fields(field.fields)

        if hasattr(field, 'choices'):
            for choice in field.choices:
                yield from flatten_fields(choice.fields)


def find_field(fieldsets, id):
    id = as_internal_id(id or '')

    for fieldset in fieldsets:
        if fieldset.id == id:
            return fieldset

        if not fieldset.id or id.startswith(fieldset.id):
            for field in flatten_fields(fieldset.fields):
                if field.id == id:
                    return field


class Fieldset(object):
    """ Represents a parsed fieldset. """

    def __init__(self, label, fields=None):
        self.label = label if label != '...' else None
        self.fields = fields or []

    @property
    def id(self):
        return as_internal_id(self.human_id)

    @property
    def human_id(self):
        return self.label or ''

    def find_field(self, *args, **kwargs):
        return find_field((self, ), *args, **kwargs)


class Choice(object):
    """ Represents a parsed choice.

    Note: Choices may have child-fields which are meant to be shown to the
    user if the given choice was selected.

    """
    def __init__(self, key, label, selected=False, fields=None):
        self.key = key
        self.label = label
        self.selected = selected
        self.fields = fields


class Field(object):
    """ Represents a parsed field. """

    def __init__(self, label, required, parent=None, fieldset=None, **kwargs):
        self.label = label
        self.required = required
        self.parent = parent
        self.fieldset = fieldset

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def id(self):
        return as_internal_id(self.human_id)

    @cached_property
    def human_id(self):
        if self.parent:
            return '/'.join((
                self.parent.human_id,
                self.label
            ))

        if self.fieldset.human_id:
            return '/'.join((
                self.fieldset.human_id,
                self.label
            ))

        return self.label

    @classmethod
    def create(cls, field, identifier, parent=None, fieldset=None):
        return cls(
            label=identifier.label,
            required=identifier.required,
            parent=parent,
            fieldset=fieldset
        )

    def parse(self, value):
        return value


class PasswordField(Field):
    type = 'password'


class EmailField(Field):
    type = 'email'


class UrlField(Field):
    type = 'url'


class DateField(Field):
    type = 'date'

    def parse(self, value):
        # the first int in an ambiguous date is assumed to be a day
        # (since our software runs in europe first and foremost)
        return dateutil_parser.parse(value, dayfirst=True).date()


class DatetimeField(Field):
    type = 'datetime'

    def parse(self, value):
        # the first int in an ambiguous date is assumed to be a day
        # (since our software runs in europe first and foremost)
        return dateutil_parser.parse(value, dayfirst=True)


class TimeField(Field):
    type = 'time'

    def parse(self, value):
        return time(*map(int, value.split(':')))


class TextField(Field):
    type = 'text'

    @classmethod
    def create(cls, field, identifier, parent=None, fieldset=None):
        regex = field.regex and re.compile(field.regex) or None

        return cls(
            label=identifier.label,
            required=identifier.required,
            parent=parent,
            fieldset=fieldset,
            maxlength=field.length or None,
            regex=regex
        )


class TextAreaField(Field):
    type = 'textarea'

    @classmethod
    def create(cls, field, identifier, parent=None, fieldset=None):
        return cls(
            label=identifier.label,
            required=identifier.required,
            parent=parent,
            fieldset=fieldset,
            rows=field.rows or None
        )


class StdnumField(Field):
    type = 'stdnum'

    @classmethod
    def create(cls, field, identifier, parent=None, fieldset=None):
        return cls(
            label=identifier.label,
            required=identifier.required,
            parent=parent,
            fieldset=fieldset,
            format=field.format
        )


class RangeField(object):

    @classmethod
    def create(cls, field, identifier, parent=None, fieldset=None):
        return cls(
            label=identifier.label,
            required=identifier.required,
            parent=parent,
            fieldset=fieldset,
            range=field[0]
        )


class IntegerRangeField(RangeField, Field):
    type = 'integer_range'

    def parse(self, value):
        return int(float(value))  # automatically truncates dots


class DecimalRangeField(RangeField, Field):
    type = 'decimal_range'

    def parse(self, value):
        return Decimal(value)


class FileinputField(Field):
    type = 'fileinput'

    @classmethod
    def create(cls, field, identifier, parent=None, fieldset=None):
        return cls(
            label=identifier.label,
            required=identifier.required,
            parent=parent,
            fieldset=fieldset,
            extensions=field.extensions
        )


class OptionsField(object):

    @classmethod
    def create(cls, field, identifier, parent=None, fieldset=None):
        choices = [
            Choice(
                key=c.label,
                label=c.label + format_pricing(c.pricing),
                selected=c.checked
            )
            for c in field.choices
        ]

        pricing = {c.label: c.pricing for c in field.choices if c.pricing}

        return cls(
            label=identifier.label,
            required=identifier.required,
            parent=parent,
            fieldset=fieldset,
            choices=choices,
            pricing=pricing or None
        )

    def parse(self, value):
        if isinstance(value, str):
            return [v.strip() for v in value.split(',')]

        return value


class RadioField(OptionsField, Field):
    type = 'radio'

    def parse(self, value):
        v = super().parse(value)
        return v and v[0] or None


class CheckboxField(OptionsField, Field):
    type = 'checkbox'


def parse_formcode(formcode):
    """ Takes the given formcode and returns an intermediate representation
    that can be used to generate forms or do other things.

    """
    parsed = yaml.load('\n'.join(translate_to_yaml(formcode)), CustomLoader)

    fieldsets = []
    field_classes = {cls.type: cls for cls in Field.__subclasses__()}
    used_ids = set()

    for fieldset in parsed:

        # fieldsets occur only at the top level
        label = next(k for k in fieldset.keys())
        fs = Fieldset(label)

        fs.fields = [
            parse_field_block(block, field_classes, used_ids, fs)
            for block in fieldset[label]
        ]

        fieldsets.append(fs)

    return fieldsets


def parse_field_block(field_block, field_classes,
                      used_ids, fieldset, parent=None):
    """ Takes the given parsed field block and yields the fields from it """

    key, field = next(i for i in field_block.items())

    identifier_src = key.rstrip('= ') + '='
    identifier = ELEMENTS.identifier.parseString(identifier_src)

    # add the nested options/dependencies in case of radio/checkbox buttons
    if isinstance(field, list):
        choices = [next(i for i in f.items()) for f in field]

        for choice, dependencies in choices:
            choice.dependencies = dependencies

        choices = [c[0] for c in choices]

        field = Bunch(choices=choices, type=choices[0].type)

        # make sure only one type is found (either radio or checkbox)
        types = {f.type for f in field.choices}
        assert types <= {'radio', 'checkbox'}
        assert len(types) == 1

    result = field_classes[field.type].create(
        field, identifier, parent, fieldset)

    if result.id in used_ids:
        raise errors.DuplicateLabelError(label=result.label)

    used_ids.add(result.id)

    # go through nested blocks and recursively add them
    if field.type in {'radio', 'checkbox'}:
        for ix, choice in enumerate(field.choices):
            if not choice.dependencies:
                continue

            result.choices[ix].fields = [
                parse_field_block(
                    field_block=child,
                    field_classes=field_classes,
                    used_ids=used_ids,
                    fieldset=fieldset,
                    parent=result
                )
                for child in choice.dependencies
            ]

    return result


def format_pricing(pricing):
    if not pricing:
        return ''

    return ' ({:.2f} {})'.format(pricing[0], pricing[1])


def match(expr, text):
    """ Returns true if the given parser expression matches the given text. """
    try:
        expr.parseString(text)
    except pp.ParseException:
        return False
    else:
        return True


def try_parse(expr, text):
    """ Returns the result of the given parser expression and text, or None.

    """
    try:
        return expr.parseString(text)
    except pp.ParseException:
        return None


def prepare(text):
    """ Takes the raw form source and prepares it for the translation into
    yaml.

    """

    lines = (l.rstrip() for l in text.split('\n'))
    lines = ((ix, l) for ix, l in enumerate(lines) if l)
    lines = ((ix, l) for ix, l in ensure_a_fieldset(lines))

    for line in lines:
        yield line


def ensure_a_fieldset(lines):
    """ Makes sure that the given lines all belong to a fieldset. That means
    adding an empty fieldset before all lines, if none is found first.

    """
    found_fieldset = False

    for ix, line in lines:
        if found_fieldset:
            yield ix, line
            continue

        if match(ELEMENTS.fieldset_title, line):
            found_fieldset = True
            yield ix, line
        else:
            found_fieldset = True
            yield -1, '# ...'
            yield ix, line


def translate_to_yaml(text):
    """ Takes the given form text and constructs an easier to parse yaml
    string.

    """

    lines = ((ix, l) for ix, l in prepare(text))
    expect_nested = False
    actual_fields = 0

    for ix, line in lines:

        indent = ' ' * (4 + (len(line) - len(line.lstrip())))

        # the top level are the fieldsets
        if match(ELEMENTS.fieldset_title, line):
            yield '- "{}":'.format(line.lstrip('# ').rstrip())
            expect_nested = False
            continue

        # fields are nested lists of dictionaries
        try:
            parse_result = try_parse(ELEMENTS.single_line_fields, line)
        except re.error:
            raise errors.InvalidFormSyntax(line=ix + 1)

        if parse_result is not None:
            yield '{indent}- "{identifier}": !{type} \'{definition}\''.format(
                indent=indent,
                type=parse_result.type,
                identifier=line.split('=')[0].strip(),
                definition=line.split('=')[1].strip()
            )
            expect_nested = len(indent) > 4
            actual_fields += 1
            continue

        # checkboxes/radios come without identifier
        parse_result = try_parse(ELEMENTS.boxes, line)
        if parse_result is not None:

            if not expect_nested:
                raise errors.InvalidFormSyntax(line=ix + 1)

            yield '{indent}- !{type} \'{definition}\':'.format(
                indent=indent,
                type=parse_result.type,
                definition=line.strip()
            )
            continue

        # identifiers which are alone contain nested checkboxes/radios
        if match(ELEMENTS.identifier, line):

            # this should have been matched by the single line field above
            if not line.endswith('='):
                raise errors.InvalidFormSyntax(line=ix + 1)

            yield '{indent}- "{identifier}":'.format(
                indent=indent,
                identifier=line.strip()
            )

            expect_nested = True
            actual_fields += 1
            continue

        raise errors.InvalidFormSyntax(line=ix + 1)

    if not actual_fields:
        raise errors.InvalidFormSyntax(line=ix + 1)



import pytest

from decimal import Decimal
from onegov.form import Form, errors, find_field
from onegov.form import parse_formcode, parse_form, flatten_fieldsets
from onegov.pay import Price
from textwrap import dedent
from webob.multidict import MultiDict
from wtforms import FileField
from wtforms import validators
from wtforms.fields.html5 import (
    DateField,
    DateTimeLocalField,
    EmailField,
    URLField,
)
from wtforms_components import TimeField


def test_parse_text():
    text = dedent("""
        First name * = ___
        Last name = ___
        Country = ___[50]
        Comment = ...[8]
        Zipcode = ___[4]/[0-9]*
        Currency = ___/[A-Z]{3}
    """)

    form_class = parse_form(text)
    form = form_class()

    fields = form._fields.values()
    assert len(fields) == 6

    assert form.first_name.label.text == 'First name'
    assert len(form.first_name.validators) == 1
    assert isinstance(form.first_name.validators[0], validators.DataRequired)

    assert form.last_name.label.text == 'Last name'
    assert len(form.last_name.validators) == 1
    assert isinstance(form.country.validators[0], validators.Optional)

    assert form.country.label.text == 'Country'
    assert len(form.country.validators) == 2
    assert isinstance(form.country.validators[0], validators.Optional)
    assert isinstance(form.country.validators[1], validators.Length)

    assert form.comment.label.text == 'Comment'
    assert form.comment.widget(form.comment) == (
        '<textarea id="comment" name="comment" rows="8"></textarea>')

    assert form.zipcode.label.text == 'Zipcode'
    assert len(form.zipcode.validators) == 3
    assert isinstance(form.zipcode.validators[0], validators.Optional)
    assert isinstance(form.zipcode.validators[1], validators.Length)
    assert isinstance(form.zipcode.validators[2], validators.Regexp)

    assert form.currency.label.text == 'Currency'
    assert len(form.currency.validators) == 2
    assert isinstance(form.currency.validators[0], validators.Optional)
    assert isinstance(form.currency.validators[1], validators.Regexp)


def test_regex_validation():
    form_class = parse_form("Zipcode = ___[4]/[0-9]+")

    form = form_class(MultiDict([('zipcode', 'abcd')]))
    form.validate()
    assert form.errors

    form = form_class(MultiDict([('zipcode', '12345')]))
    form.validate()
    assert form.errors

    form = form_class(MultiDict([('zipcode', '1234')]))
    form.validate()
    assert not form.errors

    form_class = parse_form("Zipcode = ___/^(yes|no)$")

    form = form_class(MultiDict([('zipcode', 'yess')]))
    form.validate()
    assert form.errors

    form = form_class(MultiDict([('zipcode', 'noes')]))
    form.validate()
    assert form.errors

    form = form_class(MultiDict([('zipcode', 'yes')]))
    form.validate()
    assert not form.errors

    form = form_class(MultiDict([('zipcode', 'no')]))
    form.validate()
    assert not form.errors


def test_parse_different_base_class():

    class Test(Form):
        foo = 'bar'

    form_class = parse_form('x = ___', Test)
    assert form_class.foo == 'bar'
    assert isinstance(form_class(), Test)


def test_unicode():
    text = dedent("""
        # Persönliche Informationen
        Bürgerort = ___
        Geschlecht =
            ( ) Männlich
            ( ) Weiblich
    """)

    form = parse_form(text)()

    assert form.personliche_informationen_burgerort.label.text == 'Bürgerort'
    assert 'Persönliche Informationen' == form.fieldsets[0].label
    assert form.personliche_informationen_geschlecht.choices == [
        ('Männlich', 'Männlich'),
        ('Weiblich', 'Weiblich')
    ]


def test_parse_fieldsets():
    text = dedent("""
        # Name
        First name = ___
        Last name = ___

        # Address
        Street = ___

        # ...
        Comment = ___
    """)

    form_class = parse_form(text)
    form = form_class()

    fields = form._fields.values()
    assert len(fields) == 4

    fieldsets = form.fieldsets
    assert len(fieldsets) == 3

    assert len(fieldsets[0]) == 2
    assert fieldsets[0].label == 'Name'
    assert fieldsets[0].is_visible
    assert fieldsets[0]['name_first_name'].label.text == 'First name'
    assert fieldsets[0]['name_last_name'].label.text == 'Last name'

    assert len(fieldsets[1]) == 1
    assert fieldsets[1].label == 'Address'
    assert fieldsets[1].is_visible
    assert fieldsets[1]['address_street'].label.text == 'Street'

    assert len(fieldsets[2]) == 1
    assert fieldsets[2].label is None
    assert not fieldsets[2].is_visible
    assert fieldsets[2]['comment'].label.text == 'Comment'


def test_fieldset_field_ids():
    text = dedent("""
        First Name = ___

        # Spouse
        First Name = ___
    """)

    form = parse_form(text)()
    hasattr(form, 'first_name')
    hasattr(form, 'spouse_first_name')


def test_duplicate_field():
    text = dedent("""
        First Name = ___
        First Name = ___
    """)

    with pytest.raises(errors.DuplicateLabelError):
        parse_form(text)


def test_dependent_field():
    text = dedent("""
        Comment =
            [ ] I have one
                Comment = ...

        # Extra
        Comment =
            [ ] I have one
                Comment = ...
    """)

    form = parse_form(text)
    assert hasattr(form, 'comment')
    assert hasattr(form, 'comment_comment')
    assert hasattr(form, 'extra_comment')
    assert hasattr(form, 'extra_comment_comment')


def test_parse_email():
    form = parse_form("E-Mail = @@@")()

    assert form.e_mail.label.text == 'E-Mail'
    assert isinstance(form.e_mail, EmailField)


def test_parse_url():
    form = parse_form("Url = http://")()

    assert form.url.label.text == 'Url'
    assert isinstance(form.url, URLField)


def test_parse_date():
    form = parse_form("Date = YYYY.MM.DD")()

    assert form.date.label.text == 'Date'
    assert isinstance(form.date, DateField)


def test_parse_datetime():
    form = parse_form("Datetime = YYYY.MM.DD HH:MM")()

    assert form.datetime.label.text == 'Datetime'
    assert isinstance(form.datetime, DateTimeLocalField)


def test_parse_time():
    form = parse_form("Time = HH:MM")()

    assert form.time.label.text == 'Time'
    assert isinstance(form.time, TimeField)


def test_parse_fileinput():
    form = parse_form("File = *.pdf|*.doc")()

    assert form.file.label.text == 'File'
    assert isinstance(form.file, FileField)


def test_parse_radio():

    text = dedent("""
        Gender =
            ( ) Male
            (x) Female
    """)

    form = parse_form(text)()

    assert len(form._fields) == 1
    assert form.gender.choices == [
        ('Male', 'Male'),
        ('Female', 'Female'),
    ]
    assert form.gender.default == 'Female'


def test_parse_checkbox():

    text = dedent("""
        Extras =
            [ ] Extra Luggage
            [x] Priority Seating
            [x] Early Boarding
    """)

    form = parse_form(text)()

    assert len(form._fields) == 1
    assert form.extras.choices == [
        ('Extra Luggage', 'Extra Luggage'),
        ('Priority Seating', 'Priority Seating'),
        ('Early Boarding', 'Early Boarding'),
    ]
    assert form.extras.default == ['Priority Seating', 'Early Boarding']


def test_parse_radio_with_pricing():

    text = dedent("""
        Drink =
            ( ) Coffee (2.50 CHF)
            (x) Tea (1.50 CHF)
    """)

    form = parse_form(text)()
    assert form.drink.pricing.rules == {
        'Coffee': Price(Decimal(2.5), 'CHF'),
        'Tea': Price(Decimal(1.5), 'CHF')
    }


def test_parse_checkbox_with_pricing():

    text = dedent("""
        Extras =
            [ ] Bacon (2.50 CHF)
            [x] Cheese (1.50 CHF)
    """)

    form = parse_form(text)()
    assert form.extras.pricing.rules == {
        'Bacon': Price(Decimal(2.5), 'CHF'),
        'Cheese': Price(Decimal(1.5), 'CHF')
    }
    assert form.extras.pricing.rules['Bacon'].amount == Decimal(2.5)
    assert form.extras.pricing.rules['Bacon'].currency == 'CHF'


def test_dependent_validation():

    text = dedent("""
        Payment * =
            ( ) Bill
                Address * = ___
            ( ) Credit Card
                Credit Card Number * = ___
    """)

    form_class = parse_form(text)
    assert len(form_class()._fields) == 3

    form = form_class(MultiDict([
        ('payment', 'Bill'),
        ('payment_address', 'Destiny Lane')
    ]))

    form.validate()
    assert not form.errors

    form = form_class(MultiDict([
        ('payment', 'Bill'),
    ]))

    form.validate()
    assert form.errors == {'payment_address': ['This field is required.']}

    form = form_class(MultiDict([
        ('payment', 'Credit Card'),
    ]))

    form.validate()
    assert form.errors == {
        'payment_credit_card_number': ['This field is required.']}

    form = form_class(MultiDict([
        ('payment', 'Credit Card'),
        ('payment_credit_card_number', '123')
    ]))

    form.validate()
    assert not form.errors

    assert 'data-depends-on="payment/Bill"' in (
        form.payment_address.widget(form.payment_address))
    assert 'data-depends-on="payment/Credit Card"' in (
        form.payment_credit_card_number.widget(
            form.payment_credit_card_number))


def test_nested_regression():

    text = dedent("""
        Delivery * =
            (x) I want it delivered
                Alternate Address =
                    (x) No
                    ( ) Yes
                        Street = ___
                        Town = ___
            ( ) I want to pick it up
        Kommentar = ...
    """)

    form_class = parse_form(text)

    form = form_class()

    assert len(form._fields) == 5
    assert len(form.delivery.choices) == 2
    assert len(form.delivery_alternate_address.choices) == 2

    assert hasattr(form, 'delivery_alternate_address_street')
    assert hasattr(form, 'delivery_alternate_address_town')


def test_stdnum_field():

    form_class = parse_form("Bank Account = # iban")
    form = form_class(MultiDict([
        ('bank_account', '')
    ]))

    form.validate()
    assert not form.errors

    form = form_class(MultiDict([
        ('bank_account', 'CH93 0076 2011 6238 5295 7')
    ]))

    form.validate()
    assert not form.errors

    form = form_class(MultiDict([
        ('bank_account', 'CH93 0000 2011 6238 5295 7')
    ]))

    form.validate()
    assert form.errors

    with pytest.raises(ImportError):
        form_class = parse_form("Invalid = # asdf")

    form_class = parse_form("ahv = # ch.ssn")
    form = form_class(MultiDict([
        ('ahv', '756.9217.0769.85')
    ]))

    form.validate()
    assert not form.errors

    form = form_class(MultiDict([
        ('ahv', '756.9217.0769.12')
    ]))

    form.validate()
    assert form.errors


def test_optional_date():

    text = "Date = YYYY.MM.DD"

    form_class = parse_form(text)

    form = form_class(MultiDict([
        ('date', '30.02.2015')
    ]))

    form.validate()
    assert form.errors == {'date': ['Not a valid date value']}

    form = form_class(MultiDict([
        ('date', '')
    ]))

    form.validate()
    assert not form.errors


def test_invalid_syntax():
    with pytest.raises(errors.InvalidFormSyntax) as e:
        parse_form(".")
    assert e.value.line == 1

    with pytest.raises(errors.InvalidFormSyntax) as e:
        parse_form("Test = __")
    assert e.value.line == 1

    with pytest.raises(errors.InvalidFormSyntax) as e:
        parse_form('\n'.join((
            "# Fields",
            "",
            "First Name *= ___",
            "Last Name * __"
        )))
    assert e.value.line == 4

    with pytest.raises(errors.InvalidFormSyntax) as e:
        parse_form('\n'.join((
            "# Fields",
            "",
            "    [x] What",
        )))
    assert e.value.line == 3

    with pytest.raises(errors.InvalidFormSyntax) as e:
        parse_form('# Personalien')
    assert e.value.line == 1

    with pytest.raises(errors.InvalidFormSyntax) as e:
        parse_form('\n'.join((
            "Ort * = ___",
            "(x) 6. Klasse",
            "(x) 7. Klasse",
            "(x) 8. Klasse",
            "(x) 9. Klasse",
        )))
    assert e.value.line == 2


def test_parse_formcode():
    fieldsets = parse_formcode("""
        # General
        First Name *= ___
        Last Name = ___[10]

        # Order
        Products =
            [ ] Pizza
                Type =
                    (x) Default
                    ( ) Gluten-Free
            [x] Burger
    """)

    assert len(fieldsets) == 2
    assert fieldsets[0].label == 'General'

    assert fieldsets[0].fields[0].type == 'text'
    assert fieldsets[0].fields[0].required
    assert fieldsets[0].fields[0].maxlength is None
    assert fieldsets[0].fields[0].id == 'general_first_name'

    assert fieldsets[0].fields[1].type == 'text'
    assert not fieldsets[0].fields[1].required
    assert fieldsets[0].fields[1].maxlength == 10
    assert fieldsets[0].fields[1].id == 'general_last_name'

    assert fieldsets[1].label == 'Order'
    assert fieldsets[1].fields[0].type == 'checkbox'

    assert fieldsets[1].fields[0].label == 'Products'
    assert fieldsets[1].fields[0].id == 'order_products'

    assert fieldsets[1].fields[0].choices[0].key == 'Pizza'
    assert fieldsets[1].fields[0].choices[0].label == 'Pizza'
    assert not fieldsets[1].fields[0].choices[0].selected

    assert fieldsets[1].fields[0].choices[1].key == 'Burger'
    assert fieldsets[1].fields[0].choices[1].label == 'Burger'
    assert fieldsets[1].fields[0].choices[1].selected

    assert fieldsets[1].fields[0].choices[0].fields[0].label == 'Type'
    assert fieldsets[1].fields[0].choices[0].fields[0].id \
        == 'order_products_type'
    assert fieldsets[1].fields[0].choices[0].fields[0].choices[0].selected
    assert not fieldsets[1].fields[0].choices[0].fields[0].choices[1].selected
    assert fieldsets[1].fields[0].choices[0].fields[0].choices[0].label \
        == 'Default'
    assert fieldsets[1].fields[0].choices[0].fields[0].choices[1].label \
        == 'Gluten-Free'


def test_parse_formcode_duplicate_fieldname():
    with pytest.raises(errors.DuplicateLabelError):
        parse_formcode("""
            # General
            Foo *= ___
            Foo = ___[10]
        """)

    with pytest.raises(errors.DuplicateLabelError):
        parse_formcode("""
            Foo *= ___
            Foo = ___[10]
        """)

    with pytest.raises(errors.DuplicateLabelError):
        parse_formcode("""
            # General
            Sure *=
                ( ) Yes
                ( ) No
                    Foo = ___
                    Foo = ___
        """)

    with pytest.raises(errors.DuplicateLabelError):
        parse_formcode("""
            Sure *=
                ( ) Yes
                ( ) No
                    Foo = ___
                    Foo = ___
        """)


def test_flatten_fieldsets():
    fieldsets = parse_formcode("""
        # General
        First Name *= ___
        Last Name *= ___[10]

        # Order
        Products =
            [ ] Pizza
                Type =
                    (x) Default
                    ( ) Gluten-Free
            [x] Burger
    """)

    fields = list(flatten_fieldsets(fieldsets))

    assert len(fields) == 4
    assert fields[0].label == 'First Name'
    assert fields[1].label == 'Last Name'
    assert fields[2].label == 'Products'
    assert fields[3].label == 'Type'


def test_integer_range():

    form_class = parse_form("Age = 21..150")
    form = form_class(MultiDict([
        ('age', '')
    ]))

    form.validate()
    assert not form.errors

    form_class = parse_form("Age *= 21..150")
    form = form_class(MultiDict([
        ('age', '')
    ]))

    form.validate()
    assert form.errors

    form_class = parse_form("Age *= 21..150")
    form = form_class(MultiDict([
        ('age', '20')
    ]))

    form.validate()
    assert form.errors

    form_class = parse_form("Age *= 21..150")
    form = form_class(MultiDict([
        ('age', '151')
    ]))

    form.validate()
    assert form.errors

    form_class = parse_form("Age *= 21..150")
    form = form_class(MultiDict([
        ('age', '21')
    ]))

    form.validate()
    assert not form.errors


def test_decimal_range():

    form_class = parse_form("Percentage = 0.00..100.00")
    form = form_class(MultiDict([
        ('percentage', '')
    ]))

    form.validate()
    assert not form.errors

    form_class = parse_form("Percentage *= 0.00..100.00")
    form = form_class(MultiDict([
        ('percentage', '')
    ]))

    form.validate()
    assert form.errors

    form_class = parse_form("Percentage = 0.00..100.00")
    form = form_class(MultiDict([
        ('percentage', '-1')
    ]))

    form.validate()
    assert form.errors

    form_class = parse_form("Percentage = 0.00..100.00")
    form = form_class(MultiDict([
        ('percentage', '101')
    ]))

    form.validate()
    assert form.errors

    form_class = parse_form("Percentage = 0.00..100.00")
    form = form_class(MultiDict([
        ('percentage', '33.33')
    ]))

    form.validate()
    assert not form.errors


def test_field_ids():
    fs = parse_formcode("""
        First Name *= ___
        Last Name = ___[10]

        # My Order
        Products =
            [ ] Pizza
                Type =
                    (x) Default
                    ( ) Gluten-Free
            [x] Burger
    """)

    assert fs[0].fields[0].id == 'first_name'
    assert fs[0].fields[0].human_id == 'First Name'
    assert fs[0].fields[1].id == 'last_name'
    assert fs[0].fields[1].human_id == 'Last Name'
    assert fs[1].fields[0].id == 'my_order_products'
    assert fs[1].fields[0].human_id == 'My Order/Products'
    assert fs[1].fields[0].choices[0].fields[0].id == 'my_order_products_type'
    assert fs[1].fields[0].choices[0].fields[0].human_id\
        == 'My Order/Products/Type'

    assert find_field(fs, None) is fs[0]
    assert find_field(fs, 'my_order') is fs[1]
    assert find_field(fs, 'My Order') is fs[1]
    assert find_field(fs, 'first_name').id == 'first_name'
    assert find_field(fs, 'First Name').id == 'first_name'
    assert find_field(fs, 'last_name').id == 'last_name'
    assert find_field(fs, 'Last Name').id == 'last_name'
    assert find_field(fs, 'my_order_products').id == 'my_order_products'
    assert find_field(fs, 'My Order/Products').id == 'my_order_products'
    assert find_field(fs, 'my_order_products_type').id\
        == 'my_order_products_type'
    assert find_field(fs, 'My Order/Products/Type').id\
        == 'my_order_products_type'

    assert fs[0].find_field('first_name').id == 'first_name'
    assert fs[0].find_field('First Name').id == 'first_name'
    assert fs[1].find_field('first_name') is None
    assert fs[1].find_field('First Name') is None



from onegov.core.html import sanitize_html
from onegov.core.utils import binary_to_dictionary
from onegov.file.utils import IMAGE_MIME_TYPES_AND_SVG
from onegov.form.widgets import MultiCheckboxWidget
from onegov.form.widgets import OrderedMultiCheckboxWidget
from onegov.form.widgets import UploadWidget
from wtforms import FileField, SelectMultipleField, TextAreaField, widgets


class MultiCheckboxField(SelectMultipleField):

    widget = MultiCheckboxWidget()
    contains_labels = True

    def __init__(self, *args, **kwargs):
        kwargs['option_widget'] = widgets.CheckboxInput()
        super().__init__(*args, **kwargs)


class OrderedMultiCheckboxField(MultiCheckboxField):
    widget = OrderedMultiCheckboxWidget()


class UploadField(FileField):
    """ A custom file field that turns the uploaded file into a compressed
    base64 string together with the filename, size and mimetype.

    """

    widget = UploadWidget()

    @property
    def is_image(self):
        return self.data and\
            self.data.get('mimetype') in IMAGE_MIME_TYPES_AND_SVG

    def process_formdata(self, valuelist):
        # the upload widget optionally includes an action with the request,
        # indicating if the existing file should be replaced, kept or deleted
        if valuelist:
            if len(valuelist) == 2:
                action, fieldstorage = valuelist
            else:
                action = 'replace'
                fieldstorage = valuelist[0]

            if action == 'replace':
                self.data = self.process_fieldstorage(fieldstorage)
            elif action == 'delete':
                self.data = {}
            elif action == 'keep':
                pass
            else:
                raise NotImplementedError()
        else:
            self.data = {}

    def process_fieldstorage(self, fs):
        self.file = getattr(fs, 'file', getattr(fs, 'stream', None))
        self.filename = getattr(fs, 'filename', None)

        if not self.file:
            return {}

        self.file.seek(0)

        try:
            return binary_to_dictionary(self.file.read(), self.filename)
        finally:
            self.file.seek(0)


class HtmlField(TextAreaField):
    """ A textfield with html with integrated sanitation. """

    def __init__(self, *args, **kwargs):
        self.form = kwargs.get('_form')

        if 'render_kw' not in kwargs or not kwargs['render_kw'].get('class_'):
            kwargs['render_kw'] = kwargs.get('render_kw', {})
            kwargs['render_kw']['class_'] = 'editor'

        super().__init__(*args, **kwargs)

    def pre_validate(self, form):
        self.data = sanitize_html(self.data)



import pytest

from depot.manager import DepotManager
from onegov.core import Framework
from onegov.form import FormApp
from onegov_testing.utils import create_app
from pytest_localserver.http import WSGIServer


@pytest.fixture(scope='function', autouse=True)
def depot():
    DepotManager.configure('default', {
        'depot.backend': 'depot.io.memory.MemoryFileStorage'
    })

    yield DepotManager.get()

    DepotManager._clear()


@pytest.fixture(scope='function')
def form_app(request):

    # we do not support react 16 yet, as it basically requires ES6
    react = 'https://unpkg.com/react@15.6.2/dist/react-with-addons.js'
    react_dom = 'https://unpkg.com/react-dom@15.6.2/dist/react-dom.js'
    fontawesome = (
        'https://maxcdn.bootstrapcdn.com/font-awesome/4.7.0'
        '/css/font-awesome.min.css'
    )

    class TestApp(Framework, FormApp):
        pass

    class Content(object):
        pass

    @TestApp.path(path='/snippets')
    class Snippets(Content):
        html = """
            <!doctype html>
            <html>
                <head>
                    <script type="text/javascript" src="{}"></script>
                    <script type="text/javascript" src="{}"></script>
                    <link rel="stylesheet" href="{}">
                    <style>
                        .formcode-toolbar-element {{
                            height: 10px;
                            width: 10px;
                        }}
                    </style>
                </head>
                <body>
                    <div class="formcode-snippets"
                        data-source='/formcode-snippets'
                        data-target='textarea'
                    />

                    <textarea></textarea>
                </body>
            </html>
        """.format(react, react_dom, fontawesome)

    @TestApp.path(path='/registry')
    class Registry(Content):
        html = """
            <!doctype html>
            <html>
                <head></head>
                <body></body>
            </html>
        """

    @TestApp.path(path='/formcode-format')
    class FormcodeFormat(Content):
        html = """
            <!doctype html>
            <html>
                <head>
                    <script type="text/javascript" src="{}"></script>
                    <script type="text/javascript" src="{}"></script>
                    <link rel="stylesheet" href="{}">
                </head>
                <body>
                    <div id="container"></div>
                    <textarea></textarea>
                </body>
            </html>
        """.format(react, react_dom, fontawesome)

    @TestApp.path(path='/formcode-select')
    class FormcodeSelect(Content):
        html = """
            <!doctype html>
            <html>
                <head>
                    <script type="text/javascript" src="{}"></script>
                    <script type="text/javascript" src="{}"></script>
                </head>
                <body>
                    <div id="container"></div>
                    <textarea></textarea>
                </body>
            </html>
        """.format(react, react_dom)

    @TestApp.html(model=Content)
    def view_content(self, request):
        request.include('formcode')
        return self.html

    return create_app(TestApp, request, use_smtp=False)


@pytest.fixture(scope='function')
def browser(browser, form_app_url):
    browser.baseurl = form_app_url
    yield browser


@pytest.fixture(scope='function')
def form_app_url(request, form_app):
    form_app.print_exceptions = True
    server = WSGIServer(application=form_app)
    server.start()
    yield server.url
    server.stop()



""" Contains renderers to display form fields. """

import humanize

from html import escape
from onegov.form import log


__all__ = ['render_field']


class Registry(object):
    """ Keeps track of all the renderers and the types they are registered for,
    making sure each renderer is only instantiated once.

    """
    def __init__(self):
        self.renderer_map = {}

    def register_for(self, *types):
        """ Decorator to register a renderer. """
        def wrapper(renderer):
            instance = renderer()

            for type in types:
                self.renderer_map[type] = instance

            return renderer
        return wrapper

    def render(self, field):
        """ Renders the given field with the correct renderer. """
        # no point rendering empty fields
        if not field.data:
            return ''

        renderer = self.renderer_map.get(field.type)

        if renderer is None:
            log.warning('No renderer found for {}'.format(field.type))
            return ''
        else:
            return renderer(field)


registry = Registry()

# public interface
render_field = registry.render


class BaseRenderer(object):
    """ Provides utility functions for all renderers. """

    def escape(self, text):
        return escape(text, quote=True)


@registry.register_for(
    'StringField',
    'TextField',
    'TextAreaField',
    'EmailField',
    'URLField'
)
class StringFieldRenderer(BaseRenderer):
    def __call__(self, field):
        return self.escape(field.data).replace('\n', '<br>')


@registry.register_for('PasswordField')
class PasswordFieldRenderer(BaseRenderer):
    def __call__(self, field):
        return '*' * len(field.data)


@registry.register_for('DateField')
class DateFieldRenderer(BaseRenderer):
    # XXX we assume German here currently - should this change we'd have
    # to add a date format to the request and pass it here - which should
    # be doable with little work (not necessary for now)
    date_format = '%d.%m.%Y'

    def __call__(self, field):
        return field.data.strftime(self.date_format)


@registry.register_for('DateTimeLocalField')
class DateTimeLocalFieldRenderer(DateFieldRenderer):
    date_format = '%d.%m.%Y %H:%M'


@registry.register_for('TimeField')
class TimeFieldRenderer(BaseRenderer):
    def __call__(self, field):
        return '{:02d}:{:02d}'.format(field.data.hour, field.data.minute)


@registry.register_for('UploadField')
class UploadFieldRenderer(BaseRenderer):

    def __call__(self, field):
        return '{filename} ({size})'.format(
            filename=self.escape(field.data['filename']),
            size=humanize.naturalsize(field.data['size'])
        )


@registry.register_for('RadioField')
class RadioFieldRenderer(BaseRenderer):

    def __call__(self, field):
        return "✓ " + self.escape(dict(field.choices)[field.data])


@registry.register_for('MultiCheckboxField')
class MultiCheckboxFieldRenderer(BaseRenderer):

    def __call__(self, field):
        return "".join(
            "✓ " + self.escape(dict(field.choices)[value]) + '<br>'
            for value in field.data
        )[:-4]


@registry.register_for('CSRFTokenField', 'HiddenField')
class NullRenderer(BaseRenderer):
    def __call__(self, field):
        return ''


@registry.register_for('DecimalField')
class DecimalRenderer(BaseRenderer):
    def __call__(self, field):
        return '{:.2f}'.format(field.data)


@registry.register_for('IntegerField')
class IntegerRenderer(BaseRenderer):
    def __call__(self, field):
        return '{}'.format(int(field.data))



import importlib
import humanize

from mimetypes import types_map
from onegov.form import _
from onegov.form.core import with_options
from onegov.form.errors import InvalidFormSyntax, DuplicateLabelError
from stdnum.exceptions import ValidationError as StdnumValidationError
from wtforms import ValidationError


class Stdnum(object):
    """ Validates a string using any python-stdnum format.

    See `<https://github.com/arthurdejong/python-stdnum>`_.

    """
    def __init__(self, format):
        module = '.'.join(p for p in format.split('.') if p)
        self.format = importlib.import_module('stdnum.' + module)

    def __call__(self, form, field):
        # only do a check for filled out values, to check for the existance
        # of any value use DataRequired!
        if not field.data:
            return

        try:
            self.format.validate(field.data)
        except StdnumValidationError:
            raise ValidationError(field.gettext('Invalid input.'))


class FileSizeLimit(object):
    """ Makes sure an uploaded file is not bigger than the given number of
    bytes.

    Expects an :class:`onegov.form.fields.UploadField` instance.

    """

    message = _(
        "The file is too large, please provide a file smaller than {}."
    )

    def __init__(self, max_bytes):
        self.max_bytes = max_bytes

    def __call__(self, form, field):
        if field.data:
            if field.data.get('size', 0) > self.max_bytes:
                message = field.gettext(self.message).format(
                    humanize.naturalsize(self.max_bytes)
                )
                raise ValidationError(message)


class WhitelistedMimeType(object):
    """ Makes sure an uploaded file is in a whitelist of allowed mimetypes.

    Expects an :class:`onegov.form.fields.UploadField` instance.
    """

    whitelist = {
        'application/excel',
        'application/vnd.ms-excel',
        'application/msword',
        'application/pdf',
        'application/zip',
        'image/gif',
        'image/jpeg',
        'image/png',
        'image/x-ms-bmp',
        'text/plain',
    }

    message = _("Files of this type are not supported.")

    def __init__(self, whitelist=None):
        if whitelist is not None:
            self.whitelist = whitelist

    def __call__(self, form, field):
        if field.data:
            if field.data['mimetype'] not in self.whitelist:
                raise ValidationError(field.gettext(self.message))


class ExpectedExtensions(WhitelistedMimeType):
    """ Makes sure an uploaded file has one of the expected extensions. Since
    extensions are not something we can count on we look up the mimetype of
    the extension and use that to check.

    Expects an :class:`onegov.form.fields.UploadField` instance.

    Usage::

        ExpectedFileType('*')  # no check, really
        ExpectedFileType('pdf')  # makes sure the given file is a pdf
    """

    def __init__(self, extensions):
        mimetypes = set(
            types_map.get('.' + ext.lstrip('.'), None) for ext in extensions)
        super().__init__(whitelist=mimetypes)


class ValidFormDefinition(object):
    """ Makes sure the given text is a valid onegov.form definition. """

    message = _("The form could not be parsed.")
    email = _("Define at least one required e-mail field ('E-Mail * = @@@')")
    syntax = _("The syntax on line {line} is not valid.")
    duplicate = _("The field '{label}' exists more than once.")
    reserved = _("'{label}' is a reserved name. Please use a different name.")

    def __init__(self, require_email_field=True, reserved_fields=None):
        self.require_email_field = require_email_field
        self.reserved_fields = reserved_fields or set()

    def __call__(self, form, field):
        if field.data:
            # XXX circular import
            from onegov.form import parse_form

            try:
                form = parse_form(field.data)()
            except InvalidFormSyntax as e:
                field.widget = with_options(
                    field.widget, **{'data-highlight-line': e.line}
                )
                raise ValidationError(
                    field.gettext(self.syntax).format(line=e.line)
                )
            except DuplicateLabelError as e:
                raise ValidationError(
                    field.gettext(self.duplicate).format(label=e.label)
                )
            except AttributeError:
                raise ValidationError(field.gettext(self.message))
            else:
                if self.require_email_field:
                    if not form.has_required_email_field:
                        raise ValidationError(field.gettext(self.email))

                if self.reserved_fields:
                    for formfield_id, formfield in form._fields.items():
                        if formfield_id in self.reserved_fields:

                            raise ValidationError(
                                field.gettext(self.reserved).format(
                                    label=formfield.label.text
                                )
                            )
